#
# Copyright (c) 2004, 2005 Nokia
# Author: Timo Savola <tsavola@movial.fi>
#
# Licensed under GPL, see COPYING for details.
#

import sys, os, shutil, signal, grp
from os.path import join, exists, isdir, isfile, islink, basename, dirname, realpath

tools_dir = dirname(realpath(sys.argv[0]))
if tools_dir.endswith("/tools/bin"):
	scratchbox_dir = tools_dir[:-10]
else:
	scratchbox_dir = "/scratchbox"

users_dir       = join(scratchbox_dir, "users")
tools_dir       = join(scratchbox_dir, "tools", "bin")
etc_dir         = join(scratchbox_dir, "etc")
compilers_dir   = join(scratchbox_dir, "compilers")
devkits_dir     = join(scratchbox_dir, "devkits")
devicetools_dir = join(scratchbox_dir, "device_tools")

if isdir("/targets"):
	sandbox_dir = "/"
else:
	sandbox_dir = join(users_dir, os.getenv("USER"))

# Get physical directory

def join_sandbox(path):
	if path.startswith("/"):
		return join(sandbox_dir, path[1:])
	else:
		return join(sandbox_dir, path)

# Flatten a list of strings using a separator

def flatten(list, sep = ":"):
	s = ""
	if list:
		for i in list:
			if s:
				s = s + sep + i
			else:
				s = i
	return s

# Return s, or none if s is empty string

def printable(s, none = "none"):
	if s:
		return s
	else:
		return none

# Convert "none" and empty string to rep

def processable(s, rep = None):
	if not s or s == "none":
		return rep
	else:
		return s

# Look up a key from a dictionary without raising exceptions

def lookup(dictionary, key, default = None):
	if key in dictionary:
		return dictionary[key]
	else:
		return default

##
## Error

class Error (Exception):
	def __init__(self, s):
		Exception.__init__(self, s)

class ForceError (Error):
	def __init__(self, s):
		Error.__init__(self, s)

class DownloadError (Error):
	def __init__(self, s):
		Error.__init__(self, s)

class SessionError (Error):
	def __init__(self, s):
		Error.__init__(self, s)

##
## Current target

def current(throw = True):
	path = join_sandbox("/targets/links/scratchbox.config")
	if islink(path):
		conf = basename(os.readlink(path))
		pos = conf.rfind(".")
		if pos > 0:
			return conf[:pos]
	if throw:
		raise Error("No current target")

##
## Setup

def setup(target = None, compiler = None, devkits = None, cputransp = None, force = False):

	# Target

	if not target:
		target = current(None)

	# Check input values

	if compiler and compiler not in list_compilers():
		raise Error("No such compiler: " + compiler)

	if devkits:
		avail = list_devkits()
		for i in devkits:
			if i not in avail:
				raise Error("No such devkit: " + i)

		for i in xrange(len(devkits)):
			head = devkits[:i]
			name = devkits[i]
			for dep in get_devkit_deps(name):
				if dep not in head:
					raise Error("Invalid devkit list: %s must precede %s" %
						    (dep, name))


	if cputransp:
		if not exists(join_sandbox(cputransp)):
			alt = join("/scratchbox/tools/bin", cputransp)
			if not exists(join_sandbox(alt)):
				raise Error("No such CPU-transparency method: " + cputransp)
			cputransp = alt

	# Fill in old values

	oldconf = read_config(target)
	if oldconf:
		if not force:
			raise ForceError("Target already exists: " + target)

		if compiler == None:
			compiler = lookup(oldconf, "SBOX_CROSS_GCC_NAME")

		if devkits == None:
			devkits = lookup(oldconf, "SBOX_DEVKIT_NAMES", "").split(":")

		if cputransp == None:
			cputransp = lookup(oldconf, "SBOX_CPUTRANSPARENCY_METHOD")

	if not compiler:
		raise Error("No compiler specified")

	# Write config file

	compinfo = get_compiler_info(compiler)

	if cputransp == None:
		cputransp = ""

	real_prefix       = "%s-%s-" % (compinfo.subarch, compinfo.os)
	real_subst_prefix = "sbox-%s" % real_prefix

	if compinfo.arch == compinfo.subarch:
		real_prefix_list = real_prefix
	else:
		real_prefix_list = "%s:%s-%s-" % (real_prefix, compinfo.arch, compinfo.os)

	HOST_LD_ARGS = '"-rpath /scratchbox/compilers/host-gcc/lib/' + \
		       ' -L/scratchbox/compilers/host-gcc/lib"'

	if compiler == "host-gcc":
		setup_script       = ""
		default_prefix     = "host-"
		cross_prefix_list  = real_prefix_list
		cross_subst_prefix = "host-"
		cross_specs        = "host-gcc.specs"
		cross_ld_args      = HOST_LD_ARGS
	else:
		setup_script       = join(compinfo.dir, "target_setup.sh")
		default_prefix     = real_prefix
		cross_prefix_list  = real_prefix_list
		cross_subst_prefix = real_subst_prefix
		cross_specs        = "gcc.specs"
		cross_ld_args      = ""

	path = join_sandbox("/targets/%s.config" % target)
	out = file(path, "w")

	out.write("# Automatically generated configuration data. DO NOT TOUCH!\n")
	out.write("\n")
	out.write("SBOX_TARGET_NAME=%s\n" % target)
	out.write("SBOX_CPU=%s\n" % compinfo.subarch)
	out.write("SBOX_OS=%s\n" % compinfo.os)
	out.write("SBOX_CLIB=%s\n" % compinfo.clib)
	out.write("SBOX_CPUTRANSPARENCY_METHOD=%s\n" % cputransp)
	out.write("SBOX_DEVKIT_NAMES=%s\n" % flatten(devkits))
	out.write("\n")
	out.write("# DO NOT USE THESE IN YOUR MAKEFILES\n")
	out.write("\n")
	out.write("SBOX_DEFAULT_GCC_PREFIX=%s\n" % default_prefix)
	out.write("SBOX_TARGET_DIR=/targets/%s\n" % target)
	out.write("SBOX_TARGET_SETUP_SCRIPT=%s\n" % setup_script)
	out.write("\n")
	out.write("SBOX_CROSS_GCC_NAME=%s\n" % compinfo.name)
	out.write("SBOX_CROSS_GCC_PREFIX_LIST=%s\n" % cross_prefix_list)
	out.write("SBOX_CROSS_GCC_SUBST_PREFIX=%s\n" % cross_subst_prefix)
	out.write("SBOX_CROSS_GCC_SPECS_FILE=%s/%s\n" % (compinfo.dir, cross_specs))
	out.write("SBOX_CROSS_GCC_DIR=%s/bin\n" % compinfo.dir)
	out.write("SBOX_CROSS_GCC_LD_ARGS=%s\n" % cross_ld_args)
	out.write("\n")
	out.write("SBOX_HOST_GCC_NAME=host-gcc\n")
	out.write("SBOX_HOST_GCC_PREFIX_LIST=host-\n")
	out.write("SBOX_HOST_GCC_SUBST_PREFIX=host-\n")
	out.write("SBOX_HOST_GCC_SPECS_FILE=/scratchbox/compilers/host-gcc/host-gcc.specs\n")
	out.write("SBOX_HOST_GCC_DIR=/scratchbox/compilers/host-gcc/bin\n")
	out.write("SBOX_HOST_GCC_LD_ARGS=%s\n" % HOST_LD_ARGS)

##
## Install

def install(target = None, clibrary = False, etc = False, devkits = False, tools = []):

	# Target

	if not target:
		target = current()

	# Config

	conf = read_config(target)
	if not conf:
		raise Error("No such target: " + target)

	compiler = conf["SBOX_CROSS_GCC_NAME"]
	root = conf["SBOX_TARGET_DIR"]

	root = join_sandbox(root)
	if not exists(root):
		os.makedirs(root)

	# C-library

	if clibrary:
		script = lookup(conf, "SBOX_TARGET_SETUP_SCRIPT")
		if script and exists(script) and not run_script(script, root):
			raise Error("Failed to install C-library to target: " + target)

	# /etc

	if etc:
		targetdir = join(root, "etc")
		if not exists(targetdir):
			os.makedirs(targetdir)

		basedir = join(etc_dir, "target_etc_base")
		for i in os.listdir(basedir):
			shutil.copy(join(basedir, i), targetdir)

		# mtab

		mtab = join(targetdir, "mtab")
		if exists(mtab):
			os.remove(mtab)
		os.symlink("/proc/mounts", mtab)

		# passwd

		path = join(targetdir, "passwd")
		existed = exists(path)

		out = file(path, "a")

		if not existed:
			out.write("root:x:0:0:root:/root:/bin/sh\n")

		out.write(flatten([os.getenv("USER"), "x", str(os.getuid()), str(os.getgid()),
				   "", os.getenv("HOME"), "/bin/sh"]))
		out.write("\n")

		# group

		path = join(targetdir, "group")
		existed = exists(path)

		out = file(path, "a")

		if not existed:
			out.write("root:*:0:\n")

		gid = os.getgid()

		groupname = os.getenv("_SBOX_USER_GROUPNAME")
		if not groupname:
			groupname = grp.getgrgid(gid)[0]

		out.write(flatten([groupname, "x", str(gid)]))
		out.write("\n")

	# Devkits

	if devkits:
		for name in lookup(conf, "SBOX_DEVKIT_NAMES", "").split(":"):
			script = join(devkits_dir, name, "target_setup.sh")
			if exists(script) and not run_script(script, root):
				raise Error("Failed to install " + name + \
						" devkit to target: " + target)

	# Tools

	root_usr = join(root, "usr")
	for name in tools:
		copy_device_tool(name, compiler, root_usr)

# Execute script using shell

def run_script(script, root):
	sh = join(tools_dir, "bash")
	return not os.spawnv(os.P_WAIT, sh, [sh, script, root])

# Copy a device tool

def copy_device_tool(name, compiler, destroot):
	copied = False

	if exists(devicetools_dir):
		for program in os.listdir(devicetools_dir):
			if program.startswith(name + "-"):
				hier = join(devicetools_dir, program, compiler)
				if isdir(hier):
					copy_hierarchy(hier, destroot)
					copied = True

	if not copied:
		raise Error("No %s versions found for compiler: %s" % (name, compiler))

# Copy file hierarchy

def copy_hierarchy(srcroot, destroot):
	prefixlen = len(srcroot) + 1

	for info in os.walk(srcroot):
		srcdir = info[0]

		destdir = join(destroot, srcdir[prefixlen:])
		if not exists(destdir):
			os.makedirs(destdir)

		for i in info[2]:
			src = join(srcdir, i)
			dest = join(destdir, i)

			if islink(src):
				if exists(dest):
					os.remove(dest)
				os.symlink(os.readlink(src), dest)
			else:
				shutil.copy(src, dest)

##
## Rootstrap

def rootstrap(target = None, path = None, url = None):

	# Target

	if not target:
		target = current()

	# Target dir

	conf = read_config(target)
	if not conf:
		raise Error("No such target: " + target)

	root = conf["SBOX_TARGET_DIR"]

	root = join_sandbox(root)
	if not exists(root):
		os.makedirs(root)

	# Download

	if url:
		base = basename(url)
		if not base:
			raise DownloadError("Invalid rootstrap URL: " + url)

		tmp = join("/tmp", "sb-config-" + os.getenv("USER"))
		if not exists(tmp):
			os.makedirs(tmp)

		path = join(tmp, base)
		if exists(path):
			os.remove(path)

		bin = join(tools_dir, "wget")
		if os.spawnv(os.P_WAIT, bin, [bin, "-P", tmp, url]):
			if exists(path):
				os.remove(path)
			raise DownloadError("Failed to download rootstrap: " + url)

		if not exists(path):
			raise DownloadError("Failed to locate downloaded rootstrap: " + path)

	# Extract

	bin = join(tools_dir, "tar")
	if os.spawnv(os.P_WAIT, bin, [bin, "xvfz", path, "-C", root, "--exclude", "./dev/*"]):
		raise Error("Failed to extract rootstrap: " + path)

	if url:
		os.remove(path)

##
## Select

def select(target = None):

	# Target

	if not target:
		target = current()

	# Config

	conf = read_config(target)
	if not conf:
		raise Error("No such target: " + target)

	root = conf["SBOX_TARGET_DIR"]

	root = join_sandbox(root)
	if not exists(root):
		os.makedirs(root)

	compdir = join(compilers_dir, conf["SBOX_CROSS_GCC_NAME"])

	# Check login sessions

	sessions = list_sessions()

	maximum = 0
	for tty in sessions:
		if os.getpid() in sessions[tty]:
			maximum = 1
			break

	if len(sessions) > maximum:
		if maximum > 0:
			word = "other "
		else:
			word = ""

		raise SessionError("You must close your " + word + "Scratchbox sessions first")

	# Create links

	linkdir = join_sandbox("/targets/links")
	conflink = join(linkdir, "scratchbox.config")
	have_previous = islink(conflink)

	if have_previous:
		from sb import sbrsh
		sbrsh.umount(target)
		del sbrsh

		os.remove(conflink)

	if not exists(linkdir):
		os.makedirs(linkdir)

	os.symlink("/targets/%s.config" % target, conflink)

	for link in open(join(etc_dir, "rootdirs")).read().split():
		dirpath = join("/targets", target, link)

		physical_dirpath = join_sandbox(dirpath)
		if not exists(physical_dirpath):
			os.makedirs(physical_dirpath)

		linkpath = join(linkdir, link)
		if islink(linkpath):
			os.remove(linkpath)

		os.symlink(dirpath, linkpath)

	archlink = join(linkdir, "arch_tools")
	if islink(archlink):
		os.remove(archlink)
	os.symlink(join(compdir, "arch_tools"), archlink)

	# Restart

	pid = os.getenv("_SBOX_SHELL_PID")
	if have_previous and pid:
		restart = os.getenv("_SBOX_RESTART_FILE")
		if restart:
			file(restart, "w").close()

		os.kill(int(pid), signal.SIGHUP)

##
## Remove, Reset

def read_target_root(target):
	conf = read_config(target)
	if conf:
		return join_sandbox(conf["SBOX_TARGET_DIR"])
	else:
		return None

def remove(target = None, force = False):
	if not target:
		target = current()

	root = read_target_root(target)
	if not root:
		if force:
			return
		raise Error("No such target: " + target)

	if not force:
		raise ForceError("You have to force me to remove a target...")

	if exists(root):
		remove_contents(root)
		os.rmdir(root)

	sbrsh = join_sandbox("/targets/%s.sbrsh" % target)
	if exists(sbrsh):
		os.remove(sbrsh)

	os.remove(join_sandbox("/targets/%s.config" % target))

def reset(target = None, force = False):
	if not target:
		target = current()

	root = read_target_root(target)
	if not root:
		if force:
			return
		raise Error("No such target: " + target)

	if not exists(root):
		return

	if not force:
		raise ForceError("You have to force me to reset a target...")

	remove_contents(root)

	# Create root directories so that links work
	if target == current(None):
		for link in open(join(etc_dir, "rootdirs")).read().split():
			dirpath = join("/targets", target, link)

			physical_dirpath = join_sandbox(dirpath)
			if not exists(physical_dirpath):
				os.makedirs(physical_dirpath)

def remove_contents(root):
	for path, dirs, files in os.walk(root, False):
		for i in files:
			os.remove(join(path, i))

		for i in dirs:
			j = join(path, i)

			if islink(j):
				os.remove(j)
			else:
				os.rmdir(j)

##
## Config Info

class ConfigInfo:
	pass

def read_config(target = None):
	if not target:
		target = current()

	path = join_sandbox("/targets/%s.config" % target)
	if isfile(path):
		config = {}
		for line in file(path).readlines():
			pos = line.find("#")
			if pos >= 0:
				line = line[:pos]
			if line.find("=") >= 0:
				key, val = line.split("=", 1)
				config[key] = val.strip()
		return config

def read_compiler(compiler):
	path = join(compilers_dir, compiler, "compiler-name")
	if isfile(path):
		return file(path).readline().strip().split(":")

def get_target_info(target = None):
	if not target:
		target = current()

	conf = read_config(target)
	if not conf:
		return None

	info = ConfigInfo()

	try:
		info.compiler = conf["SBOX_CROSS_GCC_NAME"]
	except KeyError:
		return None

	dk = lookup(conf, "SBOX_DEVKIT_NAMES", "").strip().split(":")
	if len(dk) == 1 and not dk[0]:
		dk = []

	info.devkits = dk

	ct = lookup(conf, "SBOX_CPUTRANSPARENCY_METHOD", "").strip()
	if ct:
		if ct.startswith("/scratchbox/tools/bin/"):
			ct = ct[len("/scratchbox/tools/bin/"):]

	info.cputransp = ct

	return info

def get_compiler_info(compiler = None):
	if not compiler:
		compiler = get_target_info().compiler
		if not compiler:
			return None

	try:
		conf = read_compiler(compiler)

		info = ConfigInfo()
		info.name    = conf[0]
		info.dir     = join("/scratchbox/compilers", info.name)
		info.subarch = conf[2]
		info.os      = conf[3]
		info.clib    = conf[4]

		if len(conf) > 5:
			info.arch = conf[5]
		else:
			info.arch = info.subarch

		return info

	except:
		return None

##
## List

def list_paths(dir):
	list = []
	for file in os.listdir(dir):
		if file[0] != ".":
			list.append(join(dir, file))
	return list

def list_compilers():
	list = []
	if exists(compilers_dir):
		for name in os.listdir(compilers_dir):
			if get_compiler_info(name):
				list.append(name)
	return list

def list_devkits():
	list = []
	if exists(devkits_dir):
		for dir in list_paths(devkits_dir):
			if isdir(dir):
				list.append(basename(dir))
	return list

def get_devkit_deps(devkit):
	path = join(devkits_dir, devkit, "depends")
	if exists(path):
		return file(path).read().split()
	return []

def list_cputransps():
	path = join(etc_dir, "cputransp-methods")
	return file(path).read().split()

def list_targets():
	list = []
	for path in list_paths(join_sandbox("/targets")):
		if isfile(path) and path.endswith(".config"):
			conf = basename(path)
			list.append(conf[:conf.rindex(".")])
	return list

def list_sessions():
	map = {}
	inode = os.stat(sandbox_dir).st_ino

	for pidstr in os.listdir("/proc"):
		try:
			pid = int(pidstr)
		except ValueError:
			continue

		path = join("/proc", pidstr, "root")
		if not islink(path) or not os.access(path, os.R_OK):
			continue

		if os.stat(path).st_ino != inode:
			continue

		for fd in xrange(3):
			link = join("/proc", pidstr, "fd", str(fd))
			if not islink(link):
				continue

			tty = os.readlink(link)
			if not tty.startswith("/dev/"):
				continue

			if tty not in map:
				map[tty] = []
			map[tty].append(pid)
			break

	return map

##
## Version

def version():
	path = join(etc_dir, "scratchbox-version")
	return file(path).readline().strip()

