/*
 * Copyright (c) 2003, 2004, 2005 Nokia
 * Author: Timo Savola <tsavola@movial.fi>
 *
 * This program is licensed under GPL (see COPYING for details)
 *
 * $Id: misc_runner.c,v 1.21 2005/02/24 09:44:00 tsavola Exp $
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <time.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/stat.h>

#define CONFIG_NAME "/targets/links/scratchbox.config"

#define ERROR_CODE  126

static const char *progname = NULL;
static FILE *logfile = NULL;
static pid_t child_pid = -1;

static void error(const char *s)
{
	if (s) {
		fprintf(stderr, "%s: %s: %s\n", progname, s, strerror(errno));
	} else {
		fprintf(stderr, "%s: %s\n", progname, strerror(errno));
	}
	exit(ERROR_CODE);
}

static char *getenv_check(const char *name)
{
	char *value;

	value = getenv(name);

	if (!value || strlen(value) == 0) {
		fprintf(stderr, "%s: %s not set\n", progname, name);
		exit(ERROR_CODE);
	}

	return value;
}

static int read_line(FILE *file,
		     char *buf,
		     int size)
{
	int comm = 0;
	int c, i = 0;

	if (feof(file)) {
		return -1;
	}

	while (1) {
		c = fgetc(file);
		if (c == EOF || c == '\n') {
			break;
		}
		if (c == '#') {
			comm = 1;
		}
		if (!comm && i < size - 1) {
			*buf++ = c;
			i++;
		}
	}
	*buf = '\0';

	return i;
}

static int is_set(char *var)
{
	char *str = getenv(var);
	return (str && strcmp(str, "no") != 0);
}

static void read_config(void)
{
	FILE *file;
	char str[1024];

	file = fopen(CONFIG_NAME, "r");
	if (!file) {
		error(CONFIG_NAME);
	}

	while (read_line(file, str, sizeof (str)) >= 0) {
		if (strchr(str, '=') != NULL) {
			putenv(strdup(str));
		}
	}

	fclose(file);
}

static void print_prefix(struct timeval *tv)
{
	struct tm *t;

	t = gmtime(&tv->tv_sec);

	fprintf(logfile, "[%04d-%02d-%02d %02d:%02d:%02d %5d] ",
		1900 + t->tm_year, t->tm_mon + 1, t->tm_mday,
		t->tm_hour, t->tm_min, t->tm_sec, getpid());
}

static void print_cmd(const char *method,
		      const char *pwd,
		      char **args)
{
	int i;

	fprintf(logfile, "method: \"%s\"  pwd: \"%s\"", method, pwd);

	fprintf(logfile, "  cmd: \"");
	for (i = 0; args[i]; i++) {
		fprintf(logfile, i == 0 ? "%s" : " %s", args[i]);
	}
	fprintf(logfile, "\"\n");

	fflush(logfile);
}

static void print_time(struct timeval *start,
		       struct timeval *end)
{
	int sec, micro;

	sec = end->tv_sec - start->tv_sec;
	micro = end->tv_usec - start->tv_usec;

	if (micro < 0) {
		sec--;
		micro += 1000000;
	} else if (micro >= 1000000) {
		sec++;
		micro -= 1000000;
	}

	fprintf(logfile, "time: %d.%06d\n", sec, micro);
	fflush(logfile);
}

static void fix_args(int argc,
		     char **argv)
{
	FILE *file;
	char buf[PATH_MAX];

	if ((argc < 2 || strcmp(argv[0], argv[1]) != 0) &&
	    (argc < 3 || strcmp(argv[0], argv[2]) != 0)) {
		return;
	}

	file = fopen(argv[0], "r");
	if (!file) {
		error(argv[0]);
	}

	if (fread(buf, 1, 2, file) != 2) {
		error(argv[0]);
	}

	if (strncmp(buf, "#!", 2) == 0) {
		if (read_line(file, buf, sizeof (buf)) > 0) {
			char *cmd;
			int end;

			for (cmd = buf; isspace(*cmd); cmd++);
			for (end = 0; cmd[end] && !isspace(cmd[end]); end++);

			if (end > 0) {
				argv[0] = strndup(cmd, end);
			}
		}
	}

	fclose(file);
}

static int method_is_sbrsh(const char *method)
{
	char *path, *base;
	int retval;

	path = strdup(method);
	if (!path) {
		error(NULL);
	}

	base = basename(path);
	retval = (strstr(base, "sbrsh") == base);

	free(path);

	return retval;
}

static char **build_args(char *pwd, int argc, char **argv, char *method)
{
	char *sbrsh_target = NULL;
	char *filter, *config = NULL, **vec;
	int sbrsh, argstart, len, i;

	sbrsh = method_is_sbrsh(method);

	if (sbrsh)

	if (sbrsh && argc > 1 && !is_set("SBOX_CPUTRANSPARENCY_NOWARN")) {
		char buf[PATH_MAX], *str;
		str = realpath(argv[1], buf);
		if (str && strncmp(str, "/tmp/", 5) == 0)
			fprintf(stderr, "Warning: Trying to execute binary from /tmp directory: %s\n",
				str);
	}

	filter = getenv("SBOX_CPUTRANSPARENCY_FILTER");
	if (filter && '\0' == filter[0])
		filter = NULL;

	if (sbrsh) {
		char *sbox_target;
		struct stat buf;

		sbox_target = getenv_check("SBOX_TARGET_NAME");

		config = malloc(9 + strlen(sbox_target) + 7 + 1);
		if (!config)
			error(NULL);

		strcpy(config, "/targets/");
		strcat(config, sbox_target);
		strcat(config, ".sbrsh");

		if (stat(config, &buf) < 0) {
			free(config);

			sbrsh_target = sbox_target;

			config = getenv("_SBOX_USER_SBRSH_CONFIG");
			if (config && '\0' == config[0])
				config = NULL;
		}
	}

	argstart = 1;
	if (sbrsh) {
		if (sbrsh_target)
			argstart += 2;
		if (config)
			argstart += 2;
		argstart += 2;
	}
	if (filter)
		++argstart;

	len = argstart + (argc - 1);
	vec = calloc(len + 1, sizeof (char *));
	if (!vec)
		error(NULL);

	i = 0;
	vec[i++] = method;
	if (sbrsh) {
		if (sbrsh_target) {
			vec[i++] = "-t";
			vec[i++] = sbrsh_target;
		}
		if (config) {
			vec[i++] = "-c";
			vec[i++] = config;
		}
		vec[i++] = "-d";
		vec[i++] = pwd;
	}
	if (filter)
		vec[i++] = filter;

	for (i = 1; i < argc; ++i)
		vec[argstart + (i - 1)] = argv[i];

	fix_args(len - argstart, &vec[argstart]);

	print_cmd(method, pwd, &vec[argstart]);

	return vec;
}

static void dummy_handler(int sig)
{
}

static void kill_handler(int sig)
{
	if (child_pid > 0)
		kill(child_pid, sig);
}

int main(int argc,
	 char **argv)
{
	char *method, *logname, *pwd, **argvec;
	struct timeval start_time;
	struct sigaction dummy_action, kill_action;

	progname = argv[0];

	if (argc < 2) {
		fprintf(stderr, "Usage: %s command [args]\n", progname);
		exit(ERROR_CODE);
	}

	if (is_set("SBOX_DISABLE_CPUTRANSPARENCY")) {
		fprintf(stderr, "%s: cannot execute binary file\n", argv[1]);
		exit(ERROR_CODE);
	}

	method = getenv("SBOX_CPUTRANSPARENCY_METHOD");

	read_config();

	logname = getenv("SBOX_CPUTRANSPARENCY_LOG");
	if (logname && logname[0] == '\0') {
		logname = NULL;
	}

	if (is_set("SBOX_CPUTRANSPARENCY_VERBOSE")) {
		fprintf(stderr, "\n ### INFO ###  CPU-Transparency!"
			"  See %s for details!\n", logname);
	}

	logfile = fopen(logname, "a");
	if (!logfile) {
		error(logname);
	}

	pwd = get_current_dir_name();
	if (!pwd) {
		error(NULL);
	}

	gettimeofday(&start_time, NULL);
	print_prefix(&start_time);

	if (!method) {
		method = getenv_check("SBOX_CPUTRANSPARENCY_METHOD");
	}

	argvec = build_args(pwd, argc, argv, method);

	dummy_action.sa_handler = dummy_handler;
	sigemptyset(&dummy_action.sa_mask);
	dummy_action.sa_flags = 0;
	sigaction(SIGCHLD, &dummy_action, NULL);

	kill_action.sa_handler = kill_handler;
	sigemptyset(&kill_action.sa_mask);
	kill_action.sa_flags = 0;
	sigaction(SIGTERM, &kill_action, NULL);
	sigaction(SIGHUP,  &kill_action, NULL);
	sigaction(SIGINT,  &kill_action, NULL);

	child_pid = fork();
	if (child_pid < 0) {
		error(NULL);
	}

	if (child_pid == 0) {
		const char *preload = "LD_PRELOAD";
		if (getenv(preload)) {
			unsetenv(preload);
		}

		execvp(argvec[0], argvec);
		error(argvec[0]);
	}

	while (1) {
		int status, rc;
		struct timeval end_time;

		if (waitpid(child_pid, &status, 0) < 0) {
			if (errno == EINTR) {
				continue;
			}

			if (errno != ECHILD) {
				int e = errno;
				kill(child_pid, SIGTERM);
				errno = e;
			}

			error(NULL);
		}

		gettimeofday(&end_time, NULL);
		print_prefix(&end_time);

		rc = ERROR_CODE;

		if (WIFEXITED(status)) {
			rc = WEXITSTATUS(status);
			fprintf(logfile, "rc: %d  ", rc);
		}

		if (WIFSIGNALED(status)) {
			fprintf(logfile, "signal: \"%s\"  ",
				strsignal(WTERMSIG(status)));
		}

		print_time(&start_time, &end_time);

		return rc;
	}

	return 0;
}
